from seedboxtools import cli, config, util
from requests.exceptions import ConnectionError
import os

def main():
    parser = cli.get_uploader_parser()
    args = parser.parse_args()

    # check config availability and load configuration
    try:
        config_fobject = open(config.default_filename)
    except (IOError, OSError), e:
        util.report_error("Cannot load configuration (%s) -- run configleecher first" % (e))
        sys.exit(7)
    cfg = config.load_config(config_fobject)
    local_download_dir = cfg.general.local_download_dir
    client = config.get_client(cfg)

    # separate the wheat from the chaff
    # and when I say 'wheat' and 'chaff', I mean 'torrent files' and 'magnet links'
    is_magnet = lambda _: _.startswith("magnet:")
    is_torrent = lambda _: not is_magnet(_)

    # give all the torrents/magnets to the client
    failed = False
    for uploadable in args.torrents:
        try:
            if is_magnet(uploadable):
                client.upload_magnet_link(uploadable)
                util.report_message("%s submitted to seedbox" % uploadable)
            elif is_torrent(uploadable):
                client.upload_torrent(uploadable)
                util.report_message("%s submitted to seedbox" % os.path.basename(uploadable))
            else:
                raise ValueError("%s is not a torrent or a magnet link" % uploadable)
        except Exception as e:
            extramessage = ""
            if isinstance(e, ConnectionError):
                if e.args[0].errno == -2:
                    extramessage = "\nCheck the hostname in your seedboxtools configuration."
            util.report_error("error while uploading %s: %s%s" % (uploadable, e, extramessage))
            failed = True

    if failed:
        return 4
    else:
        return 0



'''
Configuration for the seedbox downloader
'''

import os
from iniparse import INIConfig
from iniparse.config import Undefined
from seedboxtools import clients

default_filename = os.path.expanduser("~/.torrentleecher.cfg")

def get_default_config():
    cfg = INIConfig()
    cfg.general.client = 'TransmissionClient'
    cfg.general.local_download_dir = os.path.join(os.path.expanduser("~"), "Downloads")
    cfg[cfg.general.client].hostname = ''
    cfg[cfg.general.client].torrents_dir = '/var/lib/transmission/torrents'
    cfg[cfg.general.client].incoming_dir = '/var/lib/transmission/Downloads'
    cfg[cfg.general.client].transmission_remote_path = 'transmission-remote'
    cfg[cfg.general.client].transmission_remote_user = 'admin'
    cfg[cfg.general.client].transmission_remote_password = ''
    cfg[cfg.general.client].torrentinfo_path = 'torrentinfo-console'
    return cfg

def load_config(fobject):
    cfg = INIConfig(fobject)
    return cfg

def save_config(cfgobject, fobject):
    text = str(cfgobject)
    fobject.write(text)
    fobject.flush()

def get_client(config):
    client_constructor = clients.lookup_client(config.general.client)
    args = {"local_download_dir":config.general.local_download_dir}
    client_props = getattr(config, config.general.client)
    args.update(set([ (x, getattr(client_props, x)) for x in client_props ]))
    args = dict(args)
    return client_constructor(**args)

def raw_input_default(prompt, default, choices=None):
    if callable(default):
	    try: default = default()
	    except Exception: default = ''
    if isinstance(default,Undefined):
	    default = ''
    if choices:
        choices_str = " or ".join(choices) + ", default "
    else:
        choices_str = ''
    string = prompt + " (%s%s): " % (choices_str, default)
    _ = raw_input(string)
    choice = _ if _ else default
    if choices and choice not in choices:
        print choice, "is not in the choices, please try again"
        return raw_input_default(prompt, default, choices)
    return choice

def wizard():
    try: import readline
    except ImportError: pass
    cfg = get_default_config()
    try:
        f = open(default_filename)
        cfg = load_config(f)
    except (IOError, OSError):
        print "Couldn't load configuration.  Creating a new configuration file."
    cfg.general.local_download_dir = raw_input_default(
          "Local download directory",
          cfg.general.local_download_dir,
    )
    cfg.general.client = raw_input_default(
          "Torrent server type",
          cfg.general.client,
          ["TorrentFluxClient", "TransmissionClient", "PulsedMedia"],
    )
    if cfg.general.client == 'TransmissionClient':
        cfg[cfg.general.client].hostname = raw_input_default(
              "Torrent server host name",
              cfg[cfg.general.client].hostname,
        )
        cfg[cfg.general.client].torrents_dir = raw_input_default(
              "Directory where the torrent server stores torrent files",
              cfg[cfg.general.client].torrents_dir,
        )
        cfg[cfg.general.client].incoming_dir = raw_input_default(
              "Directory where the torrent server stores downloaded files",
              cfg[cfg.general.client].incoming_dir,
        )
        cfg[cfg.general.client].transmission_remote_path = raw_input_default(
              "Command to run transmission-remote locally",
              cfg[cfg.general.client].transmission_remote_path,
        )
        cfg[cfg.general.client].transmission_remote_user = raw_input_default(
              "User name for transmission-remote",
              cfg[cfg.general.client].transmission_remote_user,
        )
        cfg[cfg.general.client].transmission_remote_password = raw_input_default(
              "Password for transmission-remote",
              cfg[cfg.general.client].transmission_remote_password,
        )
        cfg[cfg.general.client].torrentinfo_path = raw_input_default(
              "Command to run torrentinfo-console in the server",
              cfg[cfg.general.client].torrentinfo_path,
        )
    elif cfg.general.client == 'TorrentFluxClient':
        cfg[cfg.general.client].hostname = raw_input_default(
              "Torrent server host name",
              cfg[cfg.general.client].hostname,
        )
        cfg[cfg.general.client].base_dir = raw_input_default(
              "Base directory where TorrentFlux stores its .transfers directory",
              cfg[cfg.general.client].base_dir,
        )
        cfg[cfg.general.client].incoming_dir = raw_input_default(
              "Directory where where TorrentFlux stores downloaded files",
              cfg[cfg.general.client].incoming_dir,
        )
        cfg[cfg.general.client].fluxcli_path = raw_input_default(
              "Command to run fluxcli in the server",
              cfg[cfg.general.client].fluxcli_path,
        )
        cfg[cfg.general.client].torrentinfo_path = raw_input_default(
              "Command to run torrentinfo-console in the server",
              cfg[cfg.general.client].torrentinfo_path,
        )
    elif cfg.general.client == 'PulsedMedia':
        cfg[cfg.general.client].hostname = raw_input_default(
              "Hostname",
              lambda: cfg[cfg.general.client].hostname,
        )
        cfg[cfg.general.client].login = raw_input_default(
              "Login",
              lambda: cfg[cfg.general.client].login,
        )
        cfg[cfg.general.client].password = raw_input_default(
              "Password",
              lambda: cfg[cfg.general.client].password,
        )
    else:
        assert 0, "Not reached"
    print "Writing this configuration to %s" % default_filename
    print "===============8<================"
    print cfg
    print "===============>8================"
    oldumask = os.umask(0077)
    save_config(cfg, open(default_filename, "w"))
    os.umask(oldumask)
    print "Done.  Provided that you have set up SSH public key authentication to your seedbox, and you have rsync installed remotely, you are ready to run leechtorrents."




#!/usr/bin/env python

from setuptools import setup
import os

dir = os.path.dirname(__file__)
path_to_main_file = os.path.join(dir, "src/seedboxtools/__init__.py")
path_to_readme = os.path.join(dir, "README.md")
for line in open(path_to_main_file):
	if line.startswith('__version__'):
		version = line.split()[-1].strip("'").strip('"')
		break
else:
	raise ValueError, '"__version__" not found in "src/seedboxtools/__init__.py"'
readme = open(path_to_readme).read(-1)

classifiers = [
'Development Status :: 5 - Production/Stable',
'Environment :: Console',
'Environment :: No Input/Output (Daemon)',
'Intended Audience :: End Users/Desktop',
'Intended Audience :: System Administrators',
'License :: OSI Approved :: GNU General Public License (GPL)',
'Operating System :: POSIX :: Linux',
'Programming Language :: Python :: 2 :: Only',
'Programming Language :: Python :: 2.7',
'Topic :: Communications :: File Sharing',
'Topic :: Utilities',
]

setup(
	name = 'seedboxtools',
	version=version,
	description = 'A tool to automate downloading finished torrents from a seedbox',
	long_description = readme,
	author='Manuel Amador (Rudd-O)',
	author_email='rudd-o@rudd-o.com',
	license="GPL",
	url = 'http://github.com/Rudd-O/seedboxtools',
	package_dir=dict([
					("seedboxtools", "src/seedboxtools"),
					]),
	classifiers = classifiers,
	packages = ["seedboxtools"],
	scripts = ["bin/configleecher", 'bin/leechtorrents', 'bin/uploadtorrents'],
	keywords = "seedbox TorrentFlux Transmission torrents",
	requires = ["iniparse", "requests (>=0.11.1)"],
	zip_safe=False,
)



'''
This is the code in charge of downloading proper
'''

import os, signal, sys, time, traceback
from seedboxtools import util, cli, config
from seedboxtools.clients import TemporaryMalfunction

# start execution here
def download(client, remove_finished=False):
    for torrent, status, filename in client.get_files_to_download():
        # Set loop vars up
        download_lockfile = ".%s.done" % filename
        fully_downloaded = os.path.exists(download_lockfile)
        seeding = status == "Seeding"

        # If the file is completely downloaded but not to be remotely removed, skip
        if fully_downloaded and not remove_finished:
            util.report_message("%s from %s is fully downloaded, continuing to next torrent" % (filename, torrent))
            continue

        # If the remote files don't exist, skip
        util.report_message("Checking if %s from torrent %s exists on server" % (filename, torrent))
        if not client.exists_on_server(filename):
            util.report_message("%s from %s is no longer available on server, continuing to next torrent" % (filename, torrent))
            continue

        if not fully_downloaded:

            # Start download.
            util.report_message("Downloading %s from torrent %s" % (filename, torrent))
            util.mark_dir_downloading_when_it_appears(filename)
            retvalue = client.transfer(filename)
            if retvalue != 0:
                # rsync failed
                util.mark_dir_error(filename)
                if retvalue == 20:
                    util.report_error("Download of %s stopped -- rsync process interrupted" % (filename,))
                    util.report_message("Finishing by user request")
                    return 2
                elif retvalue < 0:
                    util.report_error("Download of %s failed -- rsync process killed with signal %s" % (filename, -retvalue))
                    util.report_message("Aborting")
                    return 1
                else:
                    util.report_error("Download of %s failed -- rsync process exited with return status %s" % (filename, retvalue))
                    util.report_message("Aborting")
                    return 1
            # Rsync successful
            # mark file as downloaded
            try: file(download_lockfile, "w").write("Done")
            except OSError, e:
                if e.errno != 17: raise
            # report successful download
            fully_downloaded = True
            util.mark_dir_complete(filename)
            util.report_message("Download of %s complete" % filename)

        else:

            if remove_finished:
                if seeding:
                    util.report_message("%s from %s is complete but still seeding, not removing" % (filename, torrent))
                else:
                    client.remove_remote_download(filename)
                    util.report_message("Removal of %s complete" % filename)

sighandled = False
def sighandler(signum, frame):
    global sighandled
    if not sighandled:
        util.report_message("Received signal %s" % signum)
        # temporarily immunize from signals
        oldhandler = signal.signal(signum, signal.SIG_IGN)
        os.killpg(0, signum)
        signal.signal(signum, oldhandler)
        sighandled = True

def mainloop():
    global sighandled

    parser = cli.get_parser()
    opts, args = parser.parse_args()
    util.set_verbose(not opts.quiet)

    # command line parameter checks
    if args: parser.error("This command accepts no arguments")

    if opts.run_every is not False:
        try:
            opts.run_every = int(opts.run_every)
            if opts.run_every < 1:
                raise ValueError
        except ValueError, e:
            parser.error("option --run-every must be a positive integer")

    # check config availability and load configuration
    try:
        config_fobject = open(config.default_filename)
    except (IOError, OSError), e:
        util.report_error("Cannot load configuration (%s) -- run configleecher first" % (e))
        sys.exit(7)
    cfg = config.load_config(config_fobject)
    local_download_dir = cfg.general.local_download_dir
    client = config.get_client(cfg)

    # check download dir and log file availability
    try:
        os.chdir(local_download_dir)
    except (IOError, OSError), e:
        util.report_error("Cannot change to download directory %r: %s" % (local_download_dir, e))
        sys.exit(4)

    if opts.logfile:
        try:
            file(opts.logfile, "a", 0)
        except (IOError, OSError), e:
            util.report_error("Cannot open log file %r: %s" % (opts.logfile, e))
            sys.exit(4)

    # daemonization and preparation
    if opts.daemonize:
        logfile = opts.logfile if opts.logfile else ".torrentleecher.log"
        util.daemonize(logfile)
        # everything else depends on the local_download_dir being the cwd
        os.chdir(local_download_dir)
    elif opts.logfile:
        # non-daemonizing version of the above block
        os.close(1)
        os.close(2)
        sys.stdout = file(opts.logfile, "a", 0)
        sys.stderr = sys.stderr
        os.dup2(1, 2)

    signal.signal(signal.SIGTERM, sighandler)
    signal.signal(signal.SIGINT, sighandler)

    # lockfile check
    if opts.lock:
            torrentleecher_lockfile = ".torrentleecher.lock"
            result = util.lock(torrentleecher_lockfile)
            if not result:
                util.report_error("Another process has a lock on the download directory")
                sys.exit(0)

    def do_guarded():
        try:
            return download(
               client=client,
               remove_finished=opts.remove_finished
            )
        except IOError, e:
            if e.errno == 4: pass
            else: traceback.print_exc()
            return 8
        except TemporaryMalfunction, e:
            util.report_error(str(e))
        except Exception, e:
            raise

    retvalue = 0
    if opts.run_every is False:
        util.report_message("Starting download of finished torrents")
        retvalue = do_guarded()
        util.report_message("Download of finished torrents complete")
    else:
        util.report_message("Starting daemon for download of finished torrents")
        while not sighandled:
            retvalue = do_guarded()
            if not sighandled: util.report_message("Sleeping %s seconds" % opts.run_every)
            for _ in xrange(opts.run_every):
                if not sighandled: time.sleep(1)
        util.report_message("Download of finished torrents complete")
    if sighandled: return 0
    return retvalue



#!/usr/bin/env python
"""Subprocess utilities for seedboxtools"""


# subprocess utilities

from subprocess import Popen, PIPE, STDOUT, call, check_call
import os, sys
import fcntl

def shell_quote(shellarg):
        return"'%s'" % shellarg.replace("'", r"'\''")

def getstdout(cmdline):
        p = Popen(cmdline, stdout=PIPE)
        output = p.communicate()[0]
        if p.returncode != 0: raise Exception, "Command %s return code %s" % (cmdline, p.returncode)
        return output

def getstdoutstderr(cmdline, inp=None): # return stoud and stderr in a single string object
        p = Popen(cmdline, stdin=PIPE, stdout=PIPE, stderr=STDOUT)
        output = p.communicate(inp)[0]
        if p.returncode != 0: raise Exception, "Command %s return code %s" % (cmdline, p.returncode)
        return output

def passthru(cmdline):
    return call(cmdline) # return status code, pass the outputs thru

def quote_cmdline(cmdline):
    """Quote a command line in list form for SSH usage"""
    return " ".join(shell_quote(x) for x in cmdline)

def ssh_getstdout(hostname, cmdline):
    cmd = quote_cmdline(cmdline)
    return getstdout(["ssh", "-o", "BatchMode yes", "-o", "ForwardX11 no", hostname, cmd])

def ssh_passthru(hostname, cmdline):
    cmd = quote_cmdline(cmdline)
    return passthru(["ssh", "-o", "BatchMode yes", "-o", "ForwardX11 no", hostname, cmd])

def firstcomponent(path):
    if not path: raise ValueError, "path cannot be empty: %r" % path
    oldpath = path
    while True:
        path = os.path.dirname(path)
        if not path or os.path.dirname(path) == path: break
        oldpath = path
    return oldpath

# unix process utilities

def daemonize(logfile=os.devnull):
    """Detach a process from the controlling terminal and run it in the
    background as a daemon.
    """

    pwd = os.getcwd()
    logfile = os.path.join(pwd, logfile)

    try: pid = os.fork()
    except OSError, e: raise Exception, "%s [%d]" % (e.strerror, e.errno)

    if (pid == 0):         # The first child.
        os.setsid()
        try: pid = os.fork()          # Fork a second child.
        except OSError, e: raise Exception, "%s [%d]" % (e.strerror, e.errno)

        if (pid == 0):     # The second child.
            os.chdir("/")
        else:
            # exit() or _exit()?  See below.
            os._exit(0)     # Exit parent (the first child) of the second child.
    else: os._exit(0)         # Exit parent of the first child.

    import resource                  # Resource usage information.
    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if (maxfd == resource.RLIM_INFINITY):
        maxfd = 1024

    # Iterate through and close all file descriptors.
    for f in [ sys.stderr, sys.stdout, sys.stdin ]:
        try: f.flush()
        except: pass

    for fd in range(0, 2):
        try: os.close(fd)
        except OSError: pass

    for f in [ sys.stderr, sys.stdout, sys.stdin ]:
        try: f.close()
        except: pass

    sys.stdin = file("/dev/null", "r")
    sys.stdout = file(logfile, "a", 0)
    sys.stderr = file(logfile, "a", 0)
    os.dup2(1, 2)

    return(0)

def lock(lockfile):
    global f
    try:
        fcntl.lockf(f.fileno(), fcntl.LOCK_UN)
        f.close()
    except: pass
    try:
        f = open(lockfile, 'w')
        fcntl.lockf(f.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError, e:
        if e.errno == 11: return False
        else: raise
    return True



# icon-setting utilities

from threading import Thread
import time

def set_dir_icon(filename, iconname):
    text = """[Desktop Entry]
Icon = % s
""" % iconname
    try: file(os.path.join(filename, ".directory"), "w").write(text)
    except: pass

def mark_dir_complete(filename): set_dir_icon(filename, "dialog-ok-apply.png")
def mark_dir_downloading(filename): set_dir_icon(filename, "document-open-remote.png")
def mark_dir_error(filename): set_dir_icon(filename, "dialog-cancel.png")

def mark_dir_downloading_when_it_appears(filename):
    isdir = os.path.isdir
    tiem = time.time
    sleep = time.sleep
    starttime = tiem()
    def dowatch():
            while not isdir(filename) and tiem() - starttime < 600: sleep(0.1)
            if isdir(filename): mark_dir_downloading(filename)
    tehrd = Thread(target=dowatch)
    tehrd.setDaemon(True)
    tehrd.start()

# tts utilities

def speakize(filename):
    try:
        filename, extension = os.path.splitext(filename)
        if len(extension) != 3: filename = filename + "." + extension
    except ValueError: pass
    for char in "[]{}.,-_": filename = filename.replace(char, " ")
    return filename

def speak(text):
    return passthru(["/usr/local/bin/speak-notification", "-n", "William", text])

# message reporting utilities

def which(program):
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def notify_send(message, transient=True):
    cmd = [
        "notify-send",
        "-a", os.path.basename(sys.argv[0]),
        "Seedbox tools",
        message,
    ]
    if transient:
        cmd.append("--hint=int:transient:1")
    return check_call(cmd)

_use_linux_gui = None
def use_linux_gui():
    global _use_linux_gui
    if _use_linux_gui is None:
        if os.environ.get("DISPLAY", None) and which("notify-send"):
            _use_linux_gui = True
        else:
            _use_linux_gui = False
    return _use_linux_gui

verbose = True
def set_verbose(v):
    global verbose
    verbose = v

def report_message(text):
    global verbose
    if verbose:
        if use_linux_gui():
            notify_send(text.capitalize())
        print >> sys.stderr, text

def report_error(text):
    if use_linux_gui():
        notify_send(text.capitalize(), transient=False)
    print >> sys.stderr, text



'''
Command - line helpers for seedboxtools
'''

import argparse
import optparse

def get_parser():
    '''returns the parser for the command line options'''
    parser = optparse.OptionParser()
    parser.add_option(
        "-g", '--logfile',
        help="redirect standard output and standard error to log file (relative to download directory; default %default)",
        action='store', dest='logfile', default=None,
    )
    parser.add_option(
        "-D", '--daemon',
        help="daemonize after start; useful for cron executions (combined with --lock); implies option -g .torrentleecher.log unless specified otherwise",
        action='store_true', dest='daemonize', default=False,
    )
    parser.add_option(
        "-t", '--run-every',
        help="start up and run forever, looping every X seconds; useful for systemd executions",
        action='store', dest='run_every', default=False
    )
    parser.add_option(
        "-r", '--remove-finished',
        help="remove downloaded torrents that are not seeding anymore",
        action='store_true', dest='remove_finished', default=False
    )
    parser.add_option(
        "-l", '--lock',
        help="lock working directory; useful for cron executions (combine with --daemon to prevent cron from jamming until downloads are finished)",
        action='store_true', dest='lock', default=False
    )
    parser.add_option(
        "-q", '--quiet',
        help="do not print anything, except for errors",
        action='store_true', dest='quiet', default=False
    )
    return parser

def get_uploader_parser():
    '''returns argument parser for uploader tool'''
    parser = argparse.ArgumentParser(description='Upload torrents and magnet links to seedbox.')
    parser.add_argument('torrents', metavar='TORRENT', nargs='+',
                        help='torrent file or magnet link')
    return parser



'''
Client classes for seedboxtools
'''

from functools import partial
import seedboxtools.util as util
import re
import os
import requests
import json

class SeedboxClientException(Exception): pass

class TemporaryMalfunction(SeedboxClientException): pass

class InvalidTorrent(SeedboxClientException):
    def __str__(self):
        return "invalid torrent file or magnet link: %r" % self.message

class SeedboxClient:

    def __init__(self, local_download_dir):
        self.local_download_dir = local_download_dir

    def get_finished_torrents(self):
        # This returns a series of tuples (torrentdescriptor, "Done") for every torrent that is done
        # If there is a temporary error, it raises TemporaryMalfunction
        raise NotImplementedError

    def get_file_name(self, torrentname):
        # This returns the file or path name to the torrent given a torrentdescriptor
        raise NotImplementedError

    def transfer(self, filename):
        raise NotImplementedError

    def exists_on_server(self, filename):
        raise NotImplementedError

    def remove_remote_download(self, filename):
        raise NotImplementedError

    def get_files_to_download(self):
        """Returns iterator with get_finished_torrents result"""
        torrents = self.get_finished_torrents()
        for name, status in torrents:
            yield (name, status, self.get_file_name(name))

    def upload_magnet_link(self, magnet_link):
        raise NotImplementedError

    def upload_torrent(self, torrent_path):
        raise NotImplementedError


class TorrentFluxClient(SeedboxClient):
    def __init__(self,
            local_download_dir,
            hostname,
            base_dir,
            incoming_dir,
            torrentinfo_path,
            fluxcli_path,
        ):
        SeedboxClient.__init__(self, local_download_dir)
        self.hostname = hostname
        self.base_dir = base_dir
        self.incoming_dir = incoming_dir
        self.fluxcli_path = fluxcli_path
        self.torrentinfo_path = torrentinfo_path

        self.getssh = partial(util.ssh_getstdout, hostname)
        self.passthru = partial(util.ssh_passthru, hostname)

    def get_finished_torrents(self):
        stdout = self.getssh([self.fluxcli, "transfers"])
        stdout = stdout.splitlines()[2:-5]
        stdout.reverse()
        stdout = [ re.match("^- (.+) - [0123456789\.]+ [KMG]B - (Seeding|Done)", line) for line in stdout ]
        pairs = [ (match.group(1), match.group(2)) for match in stdout if match ]
        return pairs

    def get_file_name(self, torrentname):
        fullpath = os.path.join(self.base_dir, ".transfers", torrentname)
        stdout = self.getssh(["env", "LANG=C", self.torrentinfo_path, fullpath]).splitlines()
        filenames = [ l[22:] for l in stdout if l.startswith("file name...........: ") ]
        if not len(filenames):
                _ = stdout.index("files...............:")
                # we disregard the actual filenames, we now want the dir name
                #filenames = [ l[3:] for l in stdout[filelistheader+1:] if l.startswith("   ") ]
                filenames = [ l[22:] for l in stdout if l.startswith("directory name......: ") ]
        assert len(filenames) is 1, "Wrong length of filenames: %r"%filenames
        return filenames[0]

    def transfer(self, filename):
        # need to single-quote the *path* for the purposes of the remote shell so it doesn't fail, because the path is used in the remote shell
        path = os.path.join(self.incoming_dir, filename)
        path = util.shell_quote(path)
        path = "%s:%s" % (self.hostname, path)
        opts = ["-arvzP"]
        cmdline = [ "rsync" ] + opts + [ path , self.local_download_dir ]
        returncode = util.passthru(cmdline)
        return returncode

    def exists_on_server(self, filename):
        path = os.path.join(self.incoming_dir, filename)
        cmd = ["test", "-e", path]
        returncode = self.passthru(cmd)
        if returncode == 1: return False
        elif returncode == 0: return True
        elif returncode == -2: raise IOError(4, "exists_on_server interrupted")
        else: raise AssertionError, "exists on server returned %s" % returncode

    def remove_remote_download(self, filename):
        returncode = self.passthru(["rm", "-rf", os.path.join(self.incoming_dir, filename)])
        if returncode == 0: return
        elif returncode == -2: raise IOError(4, "remove_remote_download interrupted")
        else: raise AssertionError, "remove dirs only returned %s" % returncode


class TransmissionClient(SeedboxClient):
    def __init__(self,
            local_download_dir,
            hostname,
            torrents_dir,
            incoming_dir,
            torrentinfo_path,
            transmission_remote_path,
            transmission_remote_user,
            transmission_remote_password,
        ):
        SeedboxClient.__init__(self, local_download_dir)
        self.hostname = hostname
        self.torrents_dir = torrents_dir
        self.incoming_dir = incoming_dir
        self.transmission_remote_path = transmission_remote_path
        self.transmission_remote_user = transmission_remote_user
        self.transmission_remote_password = transmission_remote_password

        self.getssh = partial(util.ssh_getstdout, hostname)
        self.passthru = partial(util.ssh_passthru, hostname)

    def get_finished_torrents(self):
        stdout = util.getstdout([
            self.transmission_remote_path,
            self.hostname,
            "--auth=%s:%s" % (self.transmission_remote_user, self.transmission_remote_password),
            "-l"
        ])
        stdout = stdout.splitlines()[1:-1]
        stdout.reverse()
        stdout = [ x.split() + [x[70:]] for x in stdout ]
        donetoseeding = lambda t: "Seeding" if t != "Stopped" else t
        stdout = [ (x[0], donetoseeding(x[8]), x[-1]) for x in stdout if x[4] in "Done" ]
        self.torrent_to_id_map = dict((x[2], x[0]) for x in stdout)
        pairs = [ (x[2], x[1]) for x in stdout]
        return pairs

    def get_file_name(self, torrentname):
        # first, cache the torrent names to IDs
        if not hasattr(self, "torrent_to_id_map"): self.get_finished_torrents()
        torrent_id = self.torrent_to_id_map[torrentname]
        stdout = util.getstdout([
            "env", "LANG=C",
            self.transmission_remote_path,
            self.hostname,
            "--auth=%s:%s" % (self.transmission_remote_user, self.transmission_remote_password),
            "-t", torrent_id,
            "-f"
        ]).splitlines()
        filename = util.firstcomponent(stdout[2][34:])
        return filename

    def transfer(self, filename):
        # need to single-quote the *path* for the purposes of the remote shell so it doesn't fail, because the path is used in the remote shell
        path = os.path.join(self.incoming_dir, filename)
        path = util.shell_quote(path)
        path = "%s:%s" % (self.hostname, path)
        opts = ["-arvzP"]
        cmdline = [ "rsync" ] + opts + [ path , self.local_download_dir ]
        returncode = util.passthru(cmdline)
        return returncode

    def exists_on_server(self, filename):
        path = os.path.join(self.incoming_dir, filename)
        cmd = ["test", "-e", path]
        returncode = self.passthru(cmd)
        if returncode == 1: return False
        elif returncode == 0: return True
        elif returncode == -2: raise IOError(4, "exists_on_server interrupted")
        else: raise AssertionError, "exists on server returned %s" % returncode

    def remove_remote_download(self, filename):
        if not hasattr(self, "torrent_to_id_map"): self.get_finished_torrents()
        if not hasattr(self, "filename_to_torrent_map"):
            self.filename_to_torrent_map = dict(
                (self.get_file_name(torrentname), torrentname) for torrentname, _ in self.get_finished_torrents()
            )
        torrent = self.filename_to_torrent_map[filename]
        torrent_id = self.torrent_to_id_map[torrent]
        returncode = util.passthru([
            "env", "LANG=C",
            self.transmission_remote_path,
            self.hostname,
            "--auth=%s:%s" % (self.transmission_remote_user, self.transmission_remote_password),
            "-t", torrent_id,
            "--remove-and-delete"
        ])
        if returncode == 0: return
        elif returncode == -2: raise IOError(4, "remove_remote_download interrupted")
        else: raise AssertionError, "remove dirs only returned %s" % returncode


class PulsedMediaClient(SeedboxClient):
    def __init__(self,
            local_download_dir,
            hostname,
            login,
            password,
        ):
        SeedboxClient.__init__(self, local_download_dir)
        self.hostname = hostname
        self.login = login
        self.password = password

        self.getssh = partial(util.ssh_getstdout, hostname)
        self.passthru = partial(util.ssh_passthru, hostname)

    def get_finished_torrents(self):
	r = requests.post(
		"https://%s/rutorrent/plugins/httprpc/action.php"%self.hostname,
		auth=(self.login, self.password),
		data="mode=list",
		verify=False,
	)
	if r.status_code == 500:
		raise TemporaryMalfunction(
			"Server is experiencing a temporary 500 status code: %s"%\
			r.content
		)
	assert r.status_code == 200, "Non-OK status code while retrieving get_finished_torrents: %r"%r.status_code
	data = json.loads(r.content)
	torrents = data["t"]
	self.torrents_cache = torrents
	try:
		self.path_for_filename_cache = dict([
			(os.path.basename(torrent[25]), torrent[25])
			for torrent in torrents.values()
		])
	except AttributeError, e:
		raise AttributeError, "normally this would be a 'list' object has no attribute 'values', but in reality something went wrong with the unserialization of JSON values, which were serialized from %r and were supposed to come from the 't' bag of JSON data -- this happens when PulsedMedia's server fucks up"%r.content
	done_torrents = []
	for key,torrent in torrents.items():
		#filename = torrent[25]
		completed_chunks = int(torrent[6])
		size_chunks = int(torrent[7])
		done = completed_chunks/size_chunks
		if done == 1: done_torrents.append( (key, "Done") )
        return done_torrents

    def get_file_name(self, torrentname):
	# in this implementation, get_finished_torrents MUST BE called first
	# or else this will bomb out with an attribute error
	torrent = self.torrents_cache[torrentname]
	return os.path.basename(torrent[25])

    def transfer(self, filename):
        # in this implementation, get_finished_torrents MUST BE called first
        # or else this will bomb out with an attribute error
        # need to single-quote the *path* for the purposes of the remote shell so it doesn't fail, because the path is used in the remote shell
        path = util.shell_quote(self.path_for_filename_cache[filename])
        path = "%s:%s" % (self.hostname, path)
        opts = ["-arvzP"]
        cmdline = [ "rsync" ] + opts + [ path , self.local_download_dir ]
        returncode = util.passthru(cmdline)
        return returncode

    def exists_on_server(self, filename):
        # in this implementation, get_finished_torrents MUST BE called first
        # or else this will bomb out with an attribute error
        path = self.path_for_filename_cache[filename]
        cmd = ["test", "-e", path] # the self.passthru takes care of quoting
        returncode = self.passthru(cmd)
        if returncode == 1: return False
        elif returncode == 0: return True
        elif returncode == -2: raise IOError(4, "exists_on_server interrupted")
        else: raise AssertionError, "exists on server returned %s" % returncode

    def upload_magnet_link(self, magnet_link):
        return self._upload(data={'url': magnet_link})

    def upload_torrent(self, torrent_path):
        files={'torrent_file': (os.path.basename(torrent_path),
                                open(torrent_path, 'rb')) }
        return self._upload(files=files)

    def _upload(self, **params):
        r = requests.post(
            "https://%s/rutorrent/php/addtorrent.php"%self.hostname,
            auth=(self.login, self.password),
            verify=False,
            **params
        )
        if r.status_code == 500:
            raise TemporaryMalfunction(
                "Server is experiencing a temporary 500 status code: %s"%\
                r.content
            )
        if 'addTorrentSuccess' in r.content:
            return
        elif 'addTorrentFailed' in r.content:
            if data in params:
                raise InvalidTorrent(params['data']['url'])
            else:
                raise InvalidTorrent(params['files']['torrent_file'][0])
        else:
            assert 0, (r.status_code, r.content)


clients = {
    'TransmissionClient':TransmissionClient,
    'TorrentFluxClient':TorrentFluxClient,
    'PulsedMedia':PulsedMediaClient,
}

def lookup_client(name):
    return clients[name]
